// This file was generated by ecm_qt_declare_logging_category(): DO NOT EDIT!

#ifndef ECM_QLOGGINGCATEGORY_LOG_PLASMA5SUPPORT_DEBUG_P_H
#define ECM_QLOGGINGCATEGORY_LOG_PLASMA5SUPPORT_DEBUG_P_H

#include <QLoggingCategory>

Q_DECLARE_LOGGING_CATEGORY(LOG_PLASMA5SUPPORT)


#endif
